from __future__ import annotations

import typing as t

import aioodbc

from . import ABCDatabaseBackend, RE_PARAM
from .common import Connection


class DBType:

    def __get__(self, obj, objtype=None):
        if obj is None:
            return 'odbc'
        breakpoint()
        return obj.options['dsn']


class Backend(ABCDatabaseBackend):

    name = 'aioodbc'
    connection_cls = Connection

    pool: t.Optional[aioodbc.Pool] = None
    db_type = 'odbc'

    def __init__(self, *args, db_type: str = None, **kwargs):
        self.db_type = db_type or self.db_type  # type: ignore
        super(Backend, self).__init__(*args, **kwargs)

    def __convert_sql__(self, sql: t.Any) -> str:
        sql = str(sql)
        if self.convert_params:
            sql = RE_PARAM.sub(r'\1?', sql)
        return sql

    async def connect(self) -> None:
        self.pool = await aioodbc.create_pool(**self.options)

    async def disconnect(self) -> None:
        pool = self.pool
        assert pool is not None, "Database is not connected"
        self.pool = None
        pool.close()
        await pool.wait_closed()

    async def acquire(self) -> aioodbc.Connection:
        pool = self.pool
        assert pool is not None, "Database is not connected"
        return await pool.acquire()

    async def release(self, conn: aioodbc.Connection):
        pool = self.pool
        assert pool is not None, "Database is not connected"
        await pool.release(conn)



from __future__ import annotations

import typing as t

import aiopg

from . import ABCDatabaseBackend
from .common import Connection as Connection_


class Connection(Connection_):

    async def _executemany(self, query: str, *params, **options) -> t.Any:
        async with self._conn.cursor() as cursor:
            for args_ in params:
                await cursor.execute(query, args_, **options)


class Backend(ABCDatabaseBackend):

    name = 'aiopg'
    db_type = 'postgresql'
    connection_cls = Connection

    pool: t.Optional[aiopg.Pool] = None

    async def connect(self) -> None:
        dsn = self.url._replace(scheme='postgresql').geturl()
        self.pool = await aiopg.create_pool(dsn, **self.options)

    async def disconnect(self) -> None:
        pool = self.pool
        assert pool is not None, "Database is not connected"
        self.pool = None
        pool.close()
        await pool.wait_closed()

    async def acquire(self) -> aiopg.Connection:
        pool = self.pool
        assert pool is not None, "Database is not connected"
        return await pool.acquire()

    async def release(self, conn: aiopg.Connection):
        pool = self.pool
        assert pool is not None, "Database is not connected"
        await pool.release(conn)



from __future__ import annotations

import trio
import trio_mysql

from . import ABCDatabaseBackend
from .common import Connection as Connection_


class Connection(Connection_):

    lock_cls = trio.Lock


class Backend(ABCDatabaseBackend):

    name = 'trio-mysql'
    db_type = 'mysql'
    connection_cls = Connection

    def __init__(self, *args, autocommit=True, charset='utf8', use_unicode=True, **options):
        """Setup default value for autocommit."""
        super(Backend, self).__init__(
            *args, autocommit=autocommit, charset=charset, use_unicode=use_unicode, **options)

    async def connect(self) -> None:
        pass

    async def disconnect(self) -> None:
        pass

    async def acquire(self) -> trio_mysql.Connection:
        conn = trio_mysql.connect(
            **self.options,
            host=self.url.hostname,
            port=self.url.port,
            user=self.url.username,
            password=self.url.password,
            db=self.url.path.strip('/'),
        )
        await conn.connect()
        return conn

    async def release(self, conn: trio_mysql.Connection):
        conn.close()



from __future__ import annotations

import typing as t

import aiomysql

from . import ABCDatabaseBackend
from .common import Connection


class Backend(ABCDatabaseBackend):

    name = 'aiomysql'
    db_type = 'mysql'
    connection_cls = Connection

    pool: t.Optional[aiomysql.Pool] = None

    async def connect(self) -> None:
        self.pool = await aiomysql.create_pool(
            **self.options,
            host=self.url.hostname,
            port=self.url.port,
            user=self.url.username,
            password=self.url.password,
            db=self.url.path.strip('/'),
        )

    async def disconnect(self) -> None:
        pool = self.pool
        assert pool is not None, "Database is not connected"
        self.pool = None
        pool.close()
        await pool.wait_closed()

    async def acquire(self) -> aiomysql.Connection:
        pool = self.pool
        assert pool is not None, "Database is not connected"
        return await pool.acquire()

    async def release(self, conn: aiomysql.Connection):
        pool = self.pool
        assert pool is not None, "Database is not connected"
        await pool.release(conn)



import typing as t

from uuid import uuid4

from . import ABCTransaction, ABCConnection
from ..record import Record


class Transaction(ABCTransaction):

    savepoint: t.Optional[str] = None

    async def _start(self) -> t.Any:
        connection = self.connection
        sql = 'BEGIN'
        if connection.transactions:
            self.savepoint = savepoint = f"AIODB_SAVEPOINT_{uuid4().hex}"
            sql = f"SAVEPOINT {savepoint}"
        connection.logger.debug((sql,))
        return await connection._execute(sql)

    async def _commit(self) -> t.Any:
        savepoint = self.savepoint
        connection = self.connection
        sql = 'COMMIT'
        if savepoint:
            sql = f"RELEASE SAVEPOINT {savepoint}"
        connection.logger.debug((sql,))
        return await connection._execute(sql)

    async def _rollback(self) -> t.Any:
        savepoint = self.savepoint
        connection = self.connection
        sql = 'ROLLBACK'
        if savepoint:
            sql = f"ROLLBACK TO SAVEPOINT {savepoint}"
        connection.logger.debug((sql,))
        return await connection._execute(sql)


class Connection(ABCConnection):

    transaction_cls = Transaction

    async def _execute(self, query: str, *params, **options) -> t.Any:
        async with self._conn.cursor() as cursor:
            await cursor.execute(query, params, **options)
            if cursor.lastrowid == 0:
                return cursor.rowcount
            return cursor.lastrowid

    async def _executemany(self, query: str, *params, **options) -> t.Any:
        async with self._conn.cursor() as cursor:
            await cursor.executemany(query, params, **options)

    async def _fetchall(self, query: str, *params, **options) -> t.List[t.Mapping]:
        async with self._conn.cursor() as cursor:
            await cursor.execute(query, params, **options)
            rows = await cursor.fetchall()
            desc = cursor.description
            return [Record(row, desc) for row in rows]

    async def _fetchmany(self, size: int, query: str, *params, **options) -> t.List[t.Mapping]:
        async with self._conn.cursor() as cursor:
            await cursor.execute(query, params, **options)
            rows = await cursor.fetchmany(size)
            desc = cursor.description
            return [Record(row, desc) for row in rows]

    async def _fetchone(self, query: str, *params, **options) -> t.Optional[t.Mapping]:
        async with self._conn.cursor() as cursor:
            await cursor.execute(query, params, **options)
            row = await cursor.fetchone()
            if row is None:
                return row
            return Record(row, cursor.description)

    async def _fetchval(self, query: str, *params, column: t.Any = 0, **options) -> t.Any:
        async with self._conn.cursor() as cursor:
            await cursor.execute(query, params, **options)
            row = await cursor.fetchone()
            if row is None:
                return row
            return row[column]

    async def _iterate(self, query: str, *params, **options) -> t.AsyncIterator[Record]:
        async with self._conn.cursor() as cursor:
            await cursor.execute(query, params, **options)
            desc = cursor.description
            while True:
                row = await cursor.fetchone()
                if row is None:
                    break
                yield Record(row, desc)


class PGReplacer:

    __slots__ = ('num',)

    def __init__(self):
        self.num = 0

    def __call__(self, match):
        self.num += 1
        return f"{match.group(1)}${self.num}"


def pg_parse_status(status: str) -> t.Union[str, int]:
    operation, params = status.split(' ', 1)
    if operation in {'INSERT'}:
        return params.split()[0]

    if operation in {'UPDATE', 'DELETE'}:
        return int(params.split()[0])

    return status



from __future__ import annotations

import typing as t

import abc
import asyncio
import logging
from re import compile as re
from urllib.parse import SplitResult, parse_qsl

from .. import logger


BACKENDS = []
SHORTCUTS = {
    'postgres': 'postgresql',
    'postgressql': 'postgresql',
    'sqllite': 'sqlite',
}
RE_PARAM = re(r'([^%])(%s)')


class ABCTransaction(abc.ABC):

    is_finished: bool = False

    def __init__(self, connection: ABCConnection):
        self.connection = connection

    @abc.abstractmethod
    async def _start(self):
        raise NotImplementedError

    @abc.abstractmethod
    async def _commit(self):
        raise NotImplementedError

    @abc.abstractmethod
    async def _rollback(self):
        raise NotImplementedError

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type: t.Type[BaseException] = None, *args):
        if not self.is_finished:
            if exc_type is not None:
                await self.rollback()
            else:
                await self.commit()

    async def start(self):
        connection = self.connection
        if not connection.is_ready:
            raise RuntimeError('There is no an acquired connection to make a transaction')

        async with connection._lock:
            await self._start()
            connection.transactions.append(self)

    async def commit(self):
        connection = self.connection
        async with connection._lock:
            await self._commit()
            connection.transactions.remove(self)
            self.is_finished = True

    async def rollback(self):
        connection = self.connection
        async with connection._lock:
            await self._rollback()
            connection.transactions.remove(self)
            self.is_finished = True


class ABCConnection(abc.ABC):

    transaction_cls: t.ClassVar[t.Type[ABCTransaction]]
    lock_cls: t.Type[asyncio.Lock] = asyncio.Lock

    __slots__ = 'database', 'logger', 'transactions', '_conn', '_lock'

    def __init__(self, database: ABCDatabaseBackend):
        self.database = database
        self.logger: logging.Logger = database.logger
        self.transactions: t.List[ABCTransaction] = []
        self._conn: t.Any = None
        self._lock: asyncio.Lock = self.lock_cls()

    @property
    def is_ready(self) -> bool:
        return self._conn is not None

    async def acquire(self) -> ABCConnection:
        if self._conn is None:
            async with self._lock:
                self._conn = await self.database.acquire()

        return self

    __aenter__ = acquire

    async def release(self, *args):
        async with self._lock:
            conn, self._conn = self._conn, None
            if conn:
                await self.database.release(conn)

    __aexit__ = release

    async def execute(self, query: t.Any, *params, **options) -> t.Any:
        sql = self.database.__convert_sql__(query)
        self.logger.debug((sql, *params))
        async with self._lock:
            return await self._execute(sql, *params, **options)

    async def executemany(self, query: t.Any, *params, **options) -> t.Any:
        sql = self.database.__convert_sql__(query)
        self.logger.debug((sql, *params))
        async with self._lock:
            return await self._executemany(sql, *params, **options)

    async def fetchall(self, query: t.Any, *params, **options) -> t.List[t.Mapping]:
        sql = self.database.__convert_sql__(query)
        self.logger.debug((sql, *params))
        async with self._lock:
            return await self._fetchall(sql, *params, **options)

    async def fetchmany(self, size: int, query: t.Any, *params, **options) -> t.List[t.Mapping]:
        sql = self.database.__convert_sql__(query)
        self.logger.debug((sql, *params))
        async with self._lock:
            return await self._fetchmany(size, sql, *params, **options)

    async def fetchone(self, query: t.Any, *params, **options) -> t.Optional[t.Mapping]:
        sql = self.database.__convert_sql__(query)
        self.logger.debug((sql, *params))
        async with self._lock:
            return await self._fetchone(sql, *params, **options)

    async def fetchval(self, query: t.Any, *params, column: t.Any = 0, **options) -> t.Any:
        sql = self.database.__convert_sql__(query)
        self.logger.debug((sql, *params))
        async with self._lock:
            return await self._fetchval(sql, *params, column=column, **options)

    async def iterate(self, query: t.Any, *params, **options) -> t.AsyncIterator:
        sql = self.database.__convert_sql__(query)
        self.logger.debug((sql, *params))
        async with self._lock:
            async for res in self._iterate(sql, *params, **options):
                yield res

    @abc.abstractmethod
    async def _execute(self, query: str, *params, **options) -> t.Any:
        raise NotImplementedError

    @abc.abstractmethod
    async def _executemany(self, query: str, *params, **options) -> t.Any:
        raise NotImplementedError

    @abc.abstractmethod
    async def _fetchall(self, query: str, *params, **options) -> t.List[t.Mapping]:
        raise NotImplementedError

    @abc.abstractmethod
    async def _fetchmany(self, size: int, query: str, *params, **options) -> t.List[t.Mapping]:
        raise NotImplementedError

    @abc.abstractmethod
    async def _fetchone(self, query: str, *params, **options) -> t.Optional[t.Mapping]:
        raise NotImplementedError

    @abc.abstractmethod
    async def _fetchval(self, query: str, *params, column: t.Any = 0, **options) -> t.Any:
        raise NotImplementedError

    @abc.abstractmethod
    async def _iterate(self, query: str, *params, **options) -> t.AsyncIterator:
        raise NotImplementedError
        yield

    def transaction(self) -> ABCTransaction:
        return self.transaction_cls(self)


class ABCDatabaseBackend(abc.ABC):

    name: t.ClassVar[str]
    db_type: t.ClassVar[str]

    connection_cls: t.ClassVar[t.Type[ABCConnection]]

    __slots__ = 'url', 'logger', 'convert_params', 'options'

    def __init__(self, url: SplitResult, logger: logging.Logger = logger,
                 convert_params: bool = False, **options):
        self.url = url
        self.logger = logger
        self.convert_params = convert_params
        self.options = dict(parse_qsl(url.query), **options)

    def __init_subclass__(cls, *args, **kwargs):
        """Register a backend."""
        BACKENDS.append(cls)
        return super().__init_subclass__(*args, **kwargs)

    def __str__(self):
        return self.db_type

    def __repr__(self):
        return f"<Backend {self}>"

    def __convert_sql__(self, sql: t.Any) -> str:
        return str(sql)

    @abc.abstractmethod
    async def connect(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    async def disconnect(self) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    async def acquire(self) -> t.Any:
        raise NotImplementedError

    @abc.abstractmethod
    async def release(self, conn: t.Any) -> None:
        raise NotImplementedError

    def connection(self) -> ABCConnection:
        return self.connection_cls(self)


#  Import available backends
#  -------------------------

from ._dummy import Backend as DummyBackend  # A dummy backend for testing  # noqa

try:
    from ._aiosqlite import Backend as AIOSQLiteBackend  # noqa
except ImportError:
    pass

try:
    from ._aiopg import Backend as AIOPGBackend  # noqa
except ImportError:
    pass

try:
    from ._asyncpg import Backend as AsyncPGBackend  # noqa
except ImportError:
    pass

try:
    from ._aiomysql import Backend as AIOMySQLBackend  # noqa
except ImportError:
    pass

try:
    from ._aioodbc import Backend as AIOODBCBackend  # noqa
except ImportError:
    pass

try:
    from ._trio_mysql import Backend as TrioMySQLBackend  # noqa
except ImportError:
    pass

try:
    from ._triopg import Backend as TrioPGBackend  # noqa
except ImportError:
    pass



from __future__ import annotations

import typing as t

import aiosqlite

from . import ABCDatabaseBackend, RE_PARAM
from .common import Connection


class Backend(ABCDatabaseBackend):

    name = 'aiosqlite'
    db_type = 'sqlite'
    connection_cls = Connection

    def __init__(self, url, isolation_level: str = None, **kwargs):
        """Set a default isolation level (enable autocommit). Fix in memory URL."""
        if ':memory:' in url.path:
            url = url._replace(path='')
        elif url.path:
            url = url._replace(path=url.path[1:])

        super(Backend, self).__init__(url, isolation_level=isolation_level, **kwargs)

    def __convert_sql__(self, sql: t.Any) -> str:
        sql = str(sql)
        if self.convert_params:
            sql = RE_PARAM.sub(r'\1?', sql)
        return sql

    async def connect(self) -> None:
        pass

    async def disconnect(self) -> None:
        pass

    async def acquire(self) -> aiosqlite.Connection:
        return await aiosqlite.connect(database=self.url.path, **self.options)

    async def release(self, conn: aiosqlite.Connection):
        await conn.commit()
        await conn.close()



from __future__ import annotations

import typing as t

import asyncpg
import trio
import trio_asyncio
import triopg

from . import ABCDatabaseBackend, ABCConnection, ABCTransaction, RE_PARAM
from .common import PGReplacer, pg_parse_status


class Transaction(ABCTransaction):

    _trans: t.Optional[asyncpg.transactions.Transaction] = None

    @property
    def trans(self) -> asyncpg.transactions.Transaction:
        if self._trans is None:
            self._trans = self.connection._conn._asyncpg_conn.transaction()
        return self._trans

    @trio_asyncio.aio_as_trio
    async def _start(self):
        return await self.trans.start()

    @trio_asyncio.aio_as_trio
    async def _commit(self):
        return await self.trans.commit()

    @trio_asyncio.aio_as_trio
    async def _rollback(self):
        return await self.trans.rollback()


class Connection(ABCConnection):

    transaction_cls = Transaction
    lock_cls = trio.Lock

    @trio_asyncio.aio_as_trio
    async def _execute(self, query: str, *params, **options) -> t.Any:
        status = await self._conn.execute(query, *params, **options)
        return pg_parse_status(status)

    @trio_asyncio.aio_as_trio
    async def _executemany(self, query: str, *params, **options) -> t.Any:
        return await self._conn.executemany(query, params, **options)

    @trio_asyncio.aio_as_trio
    async def _fetchall(self, query: str, *params, **options) -> t.List[asyncpg.Record]:
        return await self._conn.fetch(query, *params, **options)

    @trio_asyncio.aio_as_trio
    async def _fetchmany(self, size: int, query: str,
                         *params, **options) -> t.List[asyncpg.Record]:
        res = await self._conn.fetch(query, *params, **options)
        return res[:size]

    @trio_asyncio.aio_as_trio
    async def _fetchone(self, query: str, *params, **options) -> t.Optional[asyncpg.Record]:
        return await self._conn.fetchrow(query, *params, **options)

    @trio_asyncio.aio_as_trio
    async def _fetchval(self, query: str, *params, column: t.Any = 0, **options) -> t.Any:
        return await self._conn.fetchval(query, *params, column=column, **options)

    async def _iterate(self, query: str, *params, **options) -> t.AsyncIterator[asyncpg.Record]:
        conn = self._conn
        async with conn.transaction():
            async for rec in conn.cursor(query, *params):
                yield rec


class Backend(ABCDatabaseBackend):

    name = 'triopg'
    db_type = 'postgresql'
    connection_cls = Connection

    pool: t.Optional[triopg._triopg.TrioPoolProxy] = None

    def __convert_sql__(self, sql: t.Any) -> str:
        sql = str(sql)
        if self.convert_params:
            sql = RE_PARAM.sub(PGReplacer(), sql)

        return sql

    async def connect(self) -> None:
        self.pool: triopg._triopg.TrioPoolProxy = triopg.create_pool(
            **self.options,
            host=self.url.hostname,
            port=self.url.port,
            user=self.url.username,
            password=self.url.password,
            database=self.url.path.strip('/'),
        )
        await self.pool.__aenter__()

    async def disconnect(self) -> None:
        pool = self.pool
        assert pool is not None, "Database is not connected"
        self.pool = None
        await pool.close()

    @trio_asyncio.aio_as_trio
    async def acquire(self) -> triopg._triopg.TrioConnectionProxy:
        assert self.pool is not None, "Database is not connected"
        pool = self.pool._asyncpg_pool
        conn = triopg._triopg.TrioConnectionProxy()
        conn._asyncpg_conn = await pool.acquire()
        return conn

    @trio_asyncio.aio_as_trio
    async def release(self, conn: triopg._triopg.TrioConnectionProxy):
        assert self.pool is not None, "Database is not connected"
        conn = conn._asyncpg_conn
        pool = self.pool._asyncpg_pool
        await pool.release(conn)



from __future__ import annotations

import typing as t

from collections.abc import Mapping


class Record(Mapping):

    __slots__ = '_values', '_keys'

    def __init__(self, values: t.Tuple, description: t.Sequence[t.Sequence]):
        self._values = values
        self._keys = tuple(d[0] for d in description)

    def __len__(self) -> int:
        return len(self._values)

    def __getitem__(self, idx: t.Union[str, int]) -> t.Any:
        if isinstance(idx, int):
            return self._values[idx]

        for key, value in self.items():
            if key == idx:
                return value

        raise KeyError

    def __contains__(self, idx) -> bool:
        return idx in self._keys

    def __iter__(self) -> t.Iterator:
        for val in self._values:
            yield val

    def __str__(self) -> str:
        return ' '.join(f"{name}={value!r}" for name, value in self.items())
        return f"{dict(self)}"

    def __repr__(self) -> str:
        return f"<Record {self}>"

    @classmethod
    def from_dict(cls, val: t.Dict) -> Record:
        keys, values = zip(*val.items())
        return cls(values, [[name] for name in keys])

    def values(self) -> t.Tuple:  # type: ignore
        return self._values

    def keys(self) -> t.Tuple:  # type: ignore
        return self._keys

    def items(self) -> zip:  # type: ignore
        return zip(self._keys, self._values)

    def __eq__(self, obj: t.Any):
        return self._values == tuple(obj)



from __future__ import annotations

import typing as t

import logging

from contextlib import asynccontextmanager
from contextvars import ContextVar
from urllib.parse import urlsplit

from . import logger
from .backends import BACKENDS, SHORTCUTS, ABCDatabaseBackend, ABCConnection, ABCTransaction


class Database:

    url: str
    backend: ABCDatabaseBackend
    is_connected: bool = False

    def __init__(self, url: str, *, logger: logging.Logger = logger, **options):
        parsed_url = urlsplit(url)

        scheme = parsed_url.scheme
        scheme = SHORTCUTS.get(scheme, scheme)
        for backend_cls in BACKENDS:
            if backend_cls.name == scheme or backend_cls.db_type == scheme:
                break
        else:
            raise ValueError(
                f"Unsupported backend: '{scheme}', please install a required database driver")

        self.url = url
        self.logger = logger
        self.backend: ABCDatabaseBackend = backend_cls(parsed_url, logger=self.logger, **options)
        self._conn_ctx: ContextVar = ContextVar('connection', default=None)

    async def connect(self) -> Database:
        """Connect the database."""
        if not self.is_connected:
            self.logger.info(f'Database connect: {self.url}')
            await self.backend.connect()
            self.is_connected = True

        return self

    __aenter__ = connect

    async def disconnect(self, *args) -> None:
        """Disconnect the database."""
        conn = self._conn_ctx.get(None)
        if conn is not None:
            await conn.release()

        if self.is_connected:
            self.logger.info(f'Database disconnect: {self.url}')
            await self.backend.disconnect()
            self.is_connected = False

        self._conn_ctx.set(None)

    __aexit__ = disconnect

    def __create_connection__(self) -> ABCConnection:
        """Create a connection."""
        conn = self.backend.connection()
        self._conn_ctx.set(conn)
        return conn

    def __get_connection__(self, force_create: bool = True) -> t.Tuple[ABCConnection, bool]:
        """Get/create a connection from/to the current context."""
        created = force_create
        conn = self._conn_ctx.get()
        if conn is None or force_create:
            created = True
            conn = self.__create_connection__()
        return conn, created

    def connection(self, create: bool = True) -> ABCConnection:
        """Get/create a connection from/to the current context."""
        conn, _ = self.__get_connection__(create)
        return conn

    def transaction(self, create: bool = False) -> ABCTransaction:
        """Create a transaction."""
        return self.connection(create).transaction()

    async def execute(self, query: t.Any, *params, **options) -> t.Any:
        """Execute a query."""
        async with manage_connection(*self.__get_connection__(False)) as conn:
            return await conn.execute(query, *params)

    async def executemany(self, query: t.Any, *params, **options) -> t.Any:
        """Execute a query many times."""
        async with manage_connection(*self.__get_connection__(False)) as conn:
            return await conn.executemany(query, *params, **options)

    async def fetchall(self, query: t.Any, *params, **options) -> t.List[t.Mapping]:
        """Fetch all rows."""
        async with manage_connection(*self.__get_connection__(False)) as conn:
            return await conn.fetchall(query, *params, **options)

    async def fetchmany(self, size: int, query: t.Any, *params, **options) -> t.List[t.Mapping]:
        """Fetch rows."""
        async with manage_connection(*self.__get_connection__(False)) as conn:
            return await conn.fetchmany(size, query, *params, **options)

    async def fetchone(self, query: t.Any, *params, **options) -> t.Optional[t.Mapping]:
        """Fetch a row."""
        async with manage_connection(*self.__get_connection__(False)) as conn:
            return await conn.fetchone(query, *params)

    async def fetchval(self, query: t.Any, *params, column: t.Any = 0, **options) -> t.Any:
        """Fetch a value."""
        async with manage_connection(*self.__get_connection__(False)) as conn:
            return await conn.fetchval(query, *params, column=column, **options)

    async def iterate(self, query: t.Any, *params, **options) -> t.AsyncIterator:
        """Iterate through results."""
        async with manage_connection(*self.__get_connection__(False)) as conn:
            async for res in conn.iterate(query, *params, **options):
                yield res


@asynccontextmanager
async def manage_connection(connection: ABCConnection, release: bool = True):
    """Acquire/release connections."""
    await connection.acquire()
    yield connection
    if release:
        await connection.release()



import typing as t

from . import ABCDatabaseBackend, ABCConnection
from .common import Transaction


class Connection(ABCConnection):

    transaction_cls = Transaction

    async def _execute(self, query: str, *params, **options) -> t.Any:
        return None

    async def _executemany(self, query: str, *params, **options) -> t.Any:
        return None

    async def _fetchall(self, query: str, *params, **options) -> t.List[t.Mapping]:
        return []

    async def _fetchmany(self, size: int, query: str, *params, **options) -> t.List[t.Mapping]:
        return []

    async def _fetchone(self, query: str, *params, **options) -> t.Optional[t.Mapping]:
        return None

    async def _fetchval(self, query: str, *params, column: t.Any = 0, **options) -> t.Any:
        return None

    async def _iterate(self, query: str, *params, **options) -> t.AsyncIterator:
        yield None


class Backend(ABCDatabaseBackend):
    """Must not be used in production."""

    name = 'dummy'
    db_type = 'dummy'
    connection_cls = Connection

    async def connect(self):
        pass

    async def disconnect(self):
        pass

    async def acquire(self):
        pass

    async def release(self, conn):
        pass



"""Async support for various databases."""

from __future__ import annotations

import logging


__version__ = '0.7.2'


logger: logging.Logger = logging.getLogger('aio-databases')
logger.addHandler(logging.NullHandler())


from .database import Database  # noqa
